import GFFFile
import ERFFile
import NeverData
from Door import Door
from Placeable import Placeable
from Item import Item
from Creature import Creature

class Area (NeverData.NeverData):
    arePropList = {
        'Name': 'CExoLocString',
        'ChanceLightning': 'Percentage',
        'ChanceRain': 'Percentage',
        'ChanceSnow': 'Percentage',
        'DayNightCycle': 'Boolean',
        'IsNight': 'Boolean',
        'ModListenCheck': 'Integer',
        'ModSpotCheck': 'Integer',
        'MoonFogAmount': 'Integer,0-15',
        'MoonShadows': 'Boolean',
        'NoRest': 'Boolean',
        'ShadowOpacity': 'Percentage',
        'SunFogAmount': 'Integer,0-15',
        'SunShadows': 'Boolean',
        'WindPower': 'Integer,0-2',
        'OnEnter': 'ResRef,NSS',
        'OnExit': 'ResRef,NSS',
        'OnHeartbeat': 'ResRef,NSS',
        'OnUserDefined': 'ResRef,NSS',
        'MoonAmbientColor': 'BGRColour',
        'MoonDiffuseColor': 'BGRColour',
        'SunDiffuseColor': 'BGRColour',
        'SunDiffuseColor': 'BGRColour',
        'SunFogColor': 'BGRColour',
        'MoonFogColor': 'BGRColour',
        'Tag': 'CExoString',
        'PlayerVsPlayer': '2daIndex,pvpsettings.2da,strref,strref',
        'LoadScreenID': '2daIndex,loadscreens.2da,StrRef,strref,Label'        
        #'Tileset': 'ResRef,SET' #not handling resrefs not in mod yet
        }

    gitPropList = {
        'AreaProperties.AmbientSndDayVol': 'Integer,0-127',
        'AreaProperties.AmbientSndNitVol': 'Integer,0-127',
        'AreaProperties.AmbientSndDay':
        '2daIndex,ambientsound.2da,Description,strref,DisplayName',
        'AreaProperties.AmbientSndNight':
        '2daIndex,ambientsound.2da,Description,strref,DisplayName',
        'AreaProperties.EnvAudio':
        '2daIndex,soundeax.2da,Description,strref,Label',
        'AreaProperties.MusicBattle':
        '2daIndex,ambientmusic.2da,Description,strref,DisplayName',
        'AreaProperties.MusicDay':
        '2daIndex,ambientmusic.2da,Description,strref,DisplayName',        
        'AreaProperties.MusicDelay':
        '2daIndex,ambientmusic.2da,Description,strref,DisplayName',
        'AreaProperties.MusicNight':
        '2daIndex,ambientmusic.2da,Description,strref,DisplayName',
        }

    gicPropList = {}
    
    def __init__(self,erfFile,areaName):
        NeverData.NeverData.__init__(self)
        self.name = areaName
        area = erfFile.getEntryByNameAndExtension(areaName,'ARE')
        if not area:
            print "couldn't find area for ",areaName
            return None
        git = erfFile.getEntryByNameAndExtension(areaName,'GIT')
        gic = erfFile.getEntryByNameAndExtension(areaName,'GIC')
        
        self.addPropList('are',self.arePropList,erfFile.getEntryContents(area).getRoot())
        self.addPropList('gic',self.gicPropList,erfFile.getEntryContents(gic).getRoot())
        self.addPropList('git',self.gitPropList,erfFile.getEntryContents(git).getRoot())

        creatures = self.gffstructDict['git'].getInterpretedEntry('Creature List')
        self.creatureList = [Creature(creature) for creature in creatures]
        doors = self.gffstructDict['git'].getInterpretedEntry('Door List')
        self.doorList = [Door(door) for door in doors]
        placeables = self.gffstructDict['git'].getInterpretedEntry('Placeable List')
        self.placeableList = [Placeable(placeable) for placeable in placeables]
        items = self.gffstructDict['git'].getInterpretedEntry('List')
        self.itemList = [Item(item) for item in items]

    def getCreatures(self):
        return self.creatureList

    def getDoors(self):
        return self.doorList

    def getPlaceables(self):
        return self.placeableList
    
    def getItems(self):
        return self.itemList
    
    def getName(self,lang=0):
        return self.gffstructDict['are'].getInterpretedEntry('Name').getString()

    def getHeight(self):
        return self.gffstructDict['are'].getInterpretedEntry('Height')

    def getWidth(self):
        return self.gffstructDict['are'].getInterpretedEntry('Width')    




    

    
