import Numeric
from wxPython.wx import *
from wxPython.glcanvas import *
from OpenGL.GL import *
from OpenGL.GLUT import *
from OpenGL.GLU import *
import sys
import copy
import threading
import math
from sets import Set
from Module import Module
from ResourceManager import ResourceManager
import globals

SINGLESELECTIONEVENT = wxNewEventType()

def EVT_MAPSINGLESELECTION(window,function):
    '''notifies about the selection of a single object in the map'''
    window.Connect(-1,-1,SINGLESELECTIONEVENT,function)

class SingleSelectionEvent(wxPyCommandEvent):
    eventType = SINGLESELECTIONEVENT
    def __init__(self,windowID,oid):
        wxPyCommandEvent.__init__(self,self.eventType,windowID)
        self.objectID = oid

    def getSelectedId(self):
        return self.objectID
    
    def Clone(self):
        self.__class__(self.GetId(),self.objectID)

MOVEEVENT = wxNewEventType()
def EVT_MAPMOVE(window,function):
    '''notifies about movement of objects in the map'''
    window.Connect(-1,-1,MOVEEVENT,function)

class MoveEvent(wxPyCommandEvent):
    eventType = MOVEEVENT
    def __init__(self,windowID,oid,x,y):
        wxPyCommandEvent.__init__(self,self.eventType,windowID)
        self.objectID = oid
        self.x = x
        self.y = y

    def getObjectId(self):
        return self.oid

    def getX(self):
        return self.x

    def getY(self):
        return self.y

    def Clone(self):
        self.__class__(self.GetId(),self.objectID,self.x,self.y)

    
class MapWindow(wxGLCanvas):
    def __init__(self,parent):
        wxGLCanvas.__init__(self, parent, -1)
        self.init = False
        self.width = 0
        self.height = 0

        self.players = {}
        self.area = None
        self.placeables = None
        self.doors = None
        self.creatures = None
        self.lock = threading.Lock()
        self.highlight = None
        self.highlightBox = None
        self.beingDragged = None
        self.selected = []

        self.zoom = 50.0
        self.lookingAtX = 0
        self.lookingAtY = 0
        self.viewAngleSky = 50.0
        self.viewAngleFloor = 90.0

        self.angleSpeed = 3.0
        
        EVT_ERASE_BACKGROUND(self, self.OnEraseBackground)
        EVT_SIZE(self, self.OnSize)
        EVT_PAINT(self, self.OnPaint)
        EVT_LEFT_DOWN(self, self.OnMouseDown)
        EVT_LEFT_UP(self, self.OnMouseUp)
        EVT_MOTION(self, self.OnMouseMotion)
        EVT_MOUSEWHEEL(self, self.OnMouseWheel)
        EVT_KEY_DOWN(self, self.OnKeyDown)
        
    def OnEraseBackground(self, event):
        pass

    def OnSize(self, event):
        size = self.GetClientSize()
        if self.GetContext():
            self.SetCurrent()
            self.ReSizeGLScene(size.width,size.height)
        event.Skip()

    def OnPaint(self, event):
        dc = wxPaintDC(self)
        dc.ResetBoundingBox()
        self.SetCurrent()
        if not self.init:
            self.InitGL(self.GetClientSize().width,self.GetClientSize().height)
            self.init = True
        self.DrawGLScene()

    def OnMouseDown(self, evt):
        if self.highlight != None:
            self.selected = [self.highlight]
            self.beingDragged = self.getThingHit(self.highlight)
            event = SingleSelectionEvent(self.GetId(),self.beingDragged.getNevereditId())
            self.GetEventHandler().AddPendingEvent(event)
            self.dragStart = (evt.GetX(),self.height-evt.GetY())
            self.setupCamera()
            x,y,z = self.project(self.beingDragged.getX()/10.0,
                                 self.beingDragged.getY()/10.0,
                                 0.0)
            self.dragOffset = (self.dragStart[0] - x,
                               self.dragStart[1] - y)
        else:
            self.selected = []

    def OnMouseUp(self, evt):
        self.beingDragged = None

    def OnKeyDown(self,evt):
        if evt.GetKeyCode() == WXK_UP:
            self.adjustZoom(3.0)
        if evt.GetKeyCode() == WXK_DOWN:
            self.adjustZoom(-3.0)
        if evt.GetKeyCode() == 87: #'w'
            self.adjustPos(1,0)
        if evt.GetKeyCode() == 83: #'s'
            self.adjustPos(-1,0)
        if evt.GetKeyCode() == 69: #'e'
            self.adjustPos(0,-1)
        if evt.GetKeyCode() == 81: #'q'
            self.adjustPos(0,1)
        if evt.GetKeyCode() == WXK_LEFT:
            self.adjustViewAngle(self.angleSpeed,0.0)
        if evt.GetKeyCode() == WXK_RIGHT:
            self.adjustViewAngle(-self.angleSpeed,0.0)
        if evt.GetKeyCode() in (313,368): #pgdown
            self.adjustViewAngle(0.0,self.angleSpeed)
        if evt.GetKeyCode() in (312,369): #pgup
            self.adjustViewAngle(0.0,-self.angleSpeed)
        
    def OnMouseWheel(self,evt):
        self.adjustZoom(float(evt.GetWheelRotation()))

    def adjustViewAngle(self,floorAdjust, skyAdjust):
        self.viewAngleFloor += floorAdjust
        if self.viewAngleFloor > 360:
            self.viewAngleFloor -= 360
        elif self.viewAngleFloor < 0:
            self.viewAngleFloor += 360
        self.viewAngleSky += skyAdjust
        if self.viewAngleSky > 90:
            self.viewAngleSky = 89.9999
        elif self.viewAngleSky < 10:
            self.viewAngleSky = 10
        self.SetupProjection()
        self.DrawGLScene()
        
    def adjustPos(self,forwardBackward,sideways):
        if not self.area:
            return
        dx = forwardBackward * math.cos((self.viewAngleFloor)*math.pi/180.0)
        dy = forwardBackward * math.sin((self.viewAngleFloor)*math.pi/180.0)
        dx += sideways * math.cos((self.viewAngleFloor+90.0)*math.pi/180.0)
        dy += sideways * math.sin((self.viewAngleFloor+90.0)*math.pi/180.0)
        adjust = True
        testX = self.lookingAtX + dx/10.0
        if testX > self.area.getWidth():
            self.lookingAtX = self.area.getWidth()
            adjust = False
        elif testX < 0:
            self.lookingAtX = 0
            adjust = False        
        testY = self.lookingAtY + dy/10.0        
        if testY > self.area.getHeight():
            self.lookingAtY = self.area.getHeight()
            adjust = False
        elif testY < 0:
            self.lookingAtY = 0
            adjust = False
        if adjust:
            self.lookingAtX = testX
            self.lookingAtY = testY
        self.SetupProjection()
        self.DrawGLScene()
        
    def adjustZoom(self,adjustment):
        self.zoom += float(adjustment)
        if self.zoom <= 0:
            self.zoom = 1
        elif self.zoom > 300:
            self.zoom = 300
        self.SetupProjection()
        self.DrawGLScene()
        
    def getThingHit(self,name):
        if name < len(self.doors):
            return self.doors[name]
        name -= len(self.doors)
        if name < len(self.placeables):
            return self.placeables[name]
        name -= len(self.placeables)
        if name < len(self.creatures):
            return self.creatures[name]
        return None

    def selectThingById(self,id):
        for i,thing in enumerate(self.doors + self.placeables + self.creatures):
            if thing.getNevereditId() == id:
                if i not in self.selected:
                    self.highlight = None
                    self.selected = [i]
                    self.centreThing(thing)
                    self.Refresh(False)

    def centreThing(self,thing):
        self.lookingAtX = thing.getX()/10.0
        self.lookingAtY = thing.getY()/10.0
        
    def OnMouseMotion(self, evt):
        if not self.area:
            return
        if not self.beingDragged:
            glSelectBuffer(100)
            glRenderMode(GL_SELECT)
            self.SetupProjection(True,evt.GetX(),self.height-evt.GetY())
            self.DrawGLScene()
            buf = glRenderMode(GL_RENDER)
            #depthMin = 100
            for hit_record in buf:
                min_depth,max_depth,names = hit_record
                #if len(names):
                    #print names[0],min_depth,max_depth,depthMin
                #if min_depth < depthMin and len(names):
                if len(names):
                    self.setHighlightBox(names[0])
                    break
                #depthMin = min_depth
            else:
                self.highlight = None
            self.SetupProjection()
            self.DrawGLScene()
        else:
            glPushMatrix()
            self.setupCamera()
            dragX = float(evt.GetX() - self.dragOffset[0])
            dragY = float(self.height-evt.GetY() - self.dragOffset[1])
            near = Numeric.array(self.unproject(dragX,dragY,0.0))
            far = Numeric.array(self.unproject(dragX,dragY,1.0))
            N = Numeric.array([0,0,1]) #this is normal to the area plane
            P = Numeric.array([1,1,0.01]) #this is on the area plane
            u = Numeric.dot(N,P-near)/Numeric.dot(N,far-near)
            intersect = near + u * (far-near)
            x = intersect[0]
            y = intersect[1]
            if x < 0:
                x = 0
            if x > self.area.getWidth():
                x = self.area.getWidth()
            if y < 0:
                y = 0
            if y > self.area.getHeight():
                y = self.area.getHeight()                
            x *= 10.0
            y *= 10.0
            self.beingDragged.setX(x)
            self.beingDragged.setY(y)
            glPopMatrix()
            event = MoveEvent(self.GetId(),self.beingDragged.getNevereditId(),x,y)
            self.GetEventHandler().AddPendingEvent(event)

            self.DrawGLScene()
            
    def setArea(self,area):
        if self.area == area:
            return
        self.lock.acquire()
        self.selected = []
        self.highlight = None
        self.area = area
        if not self.area:
            self.selected = []
            self.placeables = None
            self.doors = None
            self.creatures = None
        else:
            self.lookingAtX = area.getWidth()/2.0
            self.lookingAtY = area.getWidth()/2.0        
            self.placeables = area.getPlaceables()
            self.doors = area.getDoors()
            self.creatures = area.getCreatures()
        self.lock.release()
        self.Refresh(True)
        
    def setPlayer(self,p):
        self.lock.acquire()
        self.players[p.playerID2] = copy.copy(p)
        self.lock.release()
        self.Refresh(False)

    def removePlayer(self,pid):
        self.lock.acquire()
        print 'removing player',pid
        del self.players[pid]
        self.lock.release()
        self.Refresh(False)

    def setHighlightBox(self,hid):
        self.highlight = hid
        class HighlightBox:
            pass
        self.highlightBox = HighlightBox()
        target = self.getThingHit(hid)
        if target:
            self.highlightBox.text = target.getName()# + ' ' + `int(target.getObjectId())`
        else:
            self.highlightBox.text = ''            
        self.highlightBox.red = 0.3
        self.highlightBox.blue = 0.3
        self.highlightBox.green = 0.3
        self.highlightBox.alpha = 0.35
        
    def output_text(self,x, y, text):
        glMatrixMode(GL_PROJECTION)
        glPushMatrix()
        glLoadIdentity()
        gluOrtho2D(0.0, self.width, 0.0, self.height)
        glMatrixMode(GL_MODELVIEW)
        glPushMatrix()
        glLoadIdentity()
        glRasterPos2f(float(x), float(y))
        for c in text:
            glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_10, ord(c))
        glPopMatrix()
        glMatrixMode(GL_PROJECTION)
        glPopMatrix()
        glMatrixMode(GL_MODELVIEW)
    
    # A general OpenGL initialization function.  Sets all of the initial parameters. 
    def InitGL(self,Width, Height):				# We call this right after our OpenGL window is created.
        self.width = Width
        self.height = Height
        glClearColor(0.0, 0.0, 0.0, 1.0)
        glClearDepth(1.0)	      # Enables Clearing Of The Depth Buffer
        glDepthFunc(GL_LESS)	      # The Type Of Depth Test To Do
        glEnable(GL_DEPTH_TEST)	      # Enables Depth Testing
        glShadeModel(GL_SMOOTH)	      # Enables Smooth Color Shading
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA)
	glInitNames()

        self.SetupProjection()

    # The function called when our window is resized
    def ReSizeGLScene(self,Width, Height):
        self.width = Width
        self.height = Height
        if Height == 0:	   # Prevent A Divide By Zero If The Window Is Too Small 
	    Height = 1
        # Reset The Current Viewport And Perspective Transformation
        glViewport(0, 0, Width, Height)
        self.SetupProjection()
        
    def SetupProjection(self,pick=False,x=0,y=0):
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        if pick:
            gluPickMatrix(x,y,5,5)
        gluPerspective(45.0, float(self.width)/float(self.height), 0.1, 100.0)
        glMatrixMode(GL_MODELVIEW)

    def project(self,x,y,z):
        vm = glGetDoublev(GL_MODELVIEW_MATRIX)
        pm = glGetDoublev(GL_PROJECTION_MATRIX)
        vp = glGetIntegerv(GL_VIEWPORT)
        return gluProject(x,y,z,vm,pm,vp)

    def unproject(self,x,y,z):
        vm = glGetDoublev(GL_MODELVIEW_MATRIX)
        pm = glGetDoublev(GL_PROJECTION_MATRIX)
        vp = glGetIntegerv(GL_VIEWPORT)
        return gluUnProject(x,y,z,vm,pm,vp)
        
    def quadricErrorCallback(self,arg):
        print gluErrorString(arg)
        
    def drawRoundedRect(self,width,height):
        cornersize = 5
        quadric = gluNewQuadric()
        gluQuadricDrawStyle(quadric,GLU_FILL)
        gluQuadricNormals(quadric,GLU_SMOOTH)
        glPushMatrix()
        glTranslatef(cornersize,cornersize,0)
        gluPartialDisk(quadric,0,cornersize,5,5,180,90)
        glTranslatef(width-2*cornersize,0,0)
        gluPartialDisk(quadric,0,cornersize,5,5,90,90)
        glTranslatef(0,height-2*cornersize,0)
        gluPartialDisk(quadric,0,cornersize,5,5,0,90)
        glTranslatef(-width+2*cornersize,0,0)
        gluPartialDisk(quadric,0,cornersize,5,5,270,90)        
        gluDeleteQuadric(quadric)
        glTranslatef(-cornersize,0,0)
        glBegin(GL_QUADS)
        glVertex3f(0,0,0)
        glVertex3f(width,0,0)
        glVertex3f(width,-height+2*cornersize,0)
        glVertex3f(0,-height+2*cornersize,0)
        glVertex3f(cornersize,cornersize,0)
        glVertex3f(width-cornersize,cornersize,0)
        glVertex3f(width-cornersize,0,0)
        glVertex3f(cornersize,0,0)
        glVertex3f(cornersize,-height+2*cornersize,0)
        glVertex3f(width-cornersize,-height+2*cornersize,0)
        glVertex3f(width-cornersize,-height+cornersize,0)
        glVertex3f(cornersize,-height+cornersize,0)
        glEnd()        
        glPopMatrix()
        
    def drawOverlays(self):
        b = self.highlightBox
        if not b:
            return
        glDisable(GL_DEPTH_TEST)
        glMatrixMode(GL_PROJECTION)
        glPushMatrix()
        glLoadIdentity()
        gluOrtho2D(0,self.width,0,self.height)
        glMatrixMode(GL_MODELVIEW)
        glPushMatrix()
        glLoadIdentity()
        glColor4f(b.red,b.green,b.blue,b.alpha)
        glTranslatef(b.x+5,b.y+5,0)
        rectWidth = glutBitmapLength(GLUT_BITMAP_TIMES_ROMAN_10,b.text)
        self.drawRoundedRect(rectWidth + 20,20)
        glColor3f(0,0,0)
        self.output_text(b.x+15,b.y+10,b.text)
        glColor3f(0.1,1,0.1)
        self.output_text(b.x+14,b.y+11,b.text)
        glPopMatrix()
        glMatrixMode(GL_PROJECTION)
        glPopMatrix()
        glMatrixMode(GL_MODELVIEW)
        glEnable(GL_DEPTH_TEST)

    def setupCamera(self):
        if not self.area:
            return
        glLoadIdentity()           # Reset The View 
        w = self.area.getWidth()
        h = self.area.getHeight()
        viewX = self.lookingAtX
        viewY = self.lookingAtY
        distance = 400.0/self.zoom
        viewZ = math.sin((180.0-self.viewAngleSky)*math.pi/180.0) * distance
        floorDistance = math.cos((180.0-self.viewAngleSky)*math.pi/180.0) * distance
        viewX += math.cos(self.viewAngleFloor*math.pi/180.0) * floorDistance
        viewY += math.sin(self.viewAngleFloor*math.pi/180.0) * floorDistance
        #print 'viewAngleSky:',self.viewAngleSky,'viewAngleFloor',self.viewAngleFloor
        #print 'viewX',viewX,'viewY',viewY,'floorDistance',floorDistance,'viewZ',viewZ
        
        gluLookAt(viewX,viewY,viewZ,
                  self.lookingAtX,self.lookingAtY,0.0,
                  0,0,1)
        
    # The main drawing function. 
    def DrawGLScene(self):
        try:
            # Clear The Screen And The Depth Buffer
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
            if not self.area:
                return
            self.setupCamera()

            w = self.area.getWidth()
            h = self.area.getHeight()


            # Draw a square (quadrilateral)
            glColor3f(0.6, 0.5, 0.4)
            glBegin(GL_QUADS)      # Start drawing a 4 sided polygon
            glVertex3f(0.0, h, 0)
            glVertex3f(w, h, 0)
            glVertex3f(w, 0, 0)
            glVertex3f(0, 0, 0)
            glEnd()

            glColor3f(0.5,0.5,0.7)
            name = 0
            glPushName(name)
            for d in self.doors:
                glLoadName(name)
                glPushMatrix()
                #print name,'values are',d.getX()/10.0,d.getY()/10.0
                #print name,'projected values would be',self.project(d.getX()/10.0,d.getY()/10.0,0)
                glTranslatef(d.getX()/10.0,d.getY()/10.0,0.01)
                if self.highlight == name:
                    self.highlightBox.x,self.highlightBox.y,self.highlightBox.z = self.project(0,0,0)
                glRotatef(d.getBearing() * 180/math.pi + 90,0,0,1)
                glRectf(0.0,0.0,0.1,0.3)
                if name in self.selected:
                    glColor4f(0.9,0.1,0.1,0.7)
                    glRectf(-0.02,-0.02,0.12,0.32)
                    glColor3f(0.5,0.5,0.7)                    
                elif self.highlight == name:
                    glColor4f(0.4,0.9,0.4,0.5)
                    glRectf(-0.02,-0.02,0.12,0.32)
                    glColor3f(0.5,0.5,0.7)
                glPopMatrix()
                name += 1

            glColor3f(0.8,0.5,0.5)
            for p in self.placeables:
                glLoadName(name)
                glPushMatrix()
                glTranslatef(p.getX()/10.0,p.getY()/10.0,0.01)
                if self.highlight == name:
                    self.highlightBox.x,self.highlightBox.y,self.highlightBox.z = self.project(0,0,0)
                glRotatef(p.getBearing() * 180/math.pi,0,0,1)
                glRectf(0.0,0.0,0.1,0.1)
                if name in self.selected:
                    glColor4f(0.9,0.1,0.1,0.7)
                    glRectf(-0.02,-0.02,0.12,0.12)
                    glColor3f(0.8,0.5,0.5)                    
                elif self.highlight == name:
                    glColor4f(0.4,0.9,0.4,0.5)
                    glRectf(-0.02,-0.02,0.12,0.12)
                    glColor3f(0.8,0.5,0.5)
                glPopMatrix()
                name += 1

            glColor3f(0.8,0.4,0.8)
            for p in self.creatures:
                glLoadName(name)
                glPushMatrix()
                glTranslatef(p.getX()/10.0,p.getY()/10.0,0.01)
                if self.highlight == name:
                    self.highlightBox.x,self.highlightBox.y,self.highlightBox.z = self.project(0,0,0)
                glRotatef(p.getBearing() * 180/math.pi,0,0,1)
                glRectf(0.0,0.0,0.1,0.1)
                if name in self.selected:
                    glColor4f(0.9,0.1,0.1,0.7)
                    glRectf(-0.02,-0.02,0.12,0.12)
                    glColor3f(0.8,0.4,0.8)                    
                elif self.highlight == name:
                    glColor4f(0.4,0.9,0.4,0.5)
                    glRectf(-0.02,-0.02,0.12,0.12)
                    glColor3f(0.8,0.4,0.8)
                glPopMatrix()
                name += 1

            for p in self.players.values():
                self.lock.acquire()
                x = p.x/10.0
                y = p.y/10.0
                a = (float(p.angle)/225.0)*360.0
                self.lock.release()
                glPushMatrix()
                glTranslatef(x,y,0.0)
                glRotate(a,0,0,1)
                glColor3f(1.0,0.4,0.4)
                glBegin(GL_LINE_LOOP)
                glVertex3f(-0.05,-0.05,0.0)
                glVertex3f(0.0,0.08,0.0)
                glVertex3f(0.05,-0.05,0.0)
                glEnd()
                glPopMatrix()
            glPopName()

            self.drawOverlays()
            self.SwapBuffers()
                
        except KeyboardInterrupt:
            print 'shutting down'
            sys.exit()

    def get_standalone(cls, mod):
        win = None
        class MyApp(wxApp):
            def OnInit(self):
                resourceManager = ResourceManager()
                resourceManager.scanGameDir('/home/pgorniak/nwn/')
                globals.resourceManager = resourceManager
                m = Module(mod)
                area = m.getArea(m.getProperty('Mod_Entry_Area')[1])
                frame = wxFrame(None, -1, "Map", wxDefaultPosition, wxSize(400,400))
                win = MapWindow(frame)
                win.setArea(area)
                frame.Show(True)
                self.SetTopWindow(frame)
                return True
        cls.app = MyApp(0)
        return win
    get_standalone = classmethod(get_standalone)

    def start_standalone(cls):
        cls.app.MainLoop()
    start_standalone = classmethod(start_standalone)
    
    
# Print message to console, and kick off the main to get it rolling.
if __name__ == "__main__":
    print "Hit ESC key to quit."
    if len(sys.argv) < 2:
        print 'usage: ' + sys.argv[0] + ' <modfile>'
        sys.exit(1)
    MapWindow.get_standalone(sys.argv[1])
    MapWindow.start_standalone()
    	
