from ERFFile import ERFFile
from GFFFile import GFFStruct
from Area import Area
from Progressor import Progressor
from NeverData import NeverData

class Module(Progressor,NeverData):
    ifoPropList = {
        "Mod_DawnHour":"Integer,0-23",
        "Mod_Description":"CExoLocString,4",
        "Mod_DuskHour":"Integer,0-23",
        "Mod_Entry_Area":"ResRef,ARE",
        "Mod_HakList":"CheckList,HAKs",
        "Mod_Hak":"CheckList,HAK",
        "Mod_Name":"CExoLocString",
        "Mod_MinPerHour": "Integer,1-255",
        "Mod_OnAcquirItem": "ResRef,NSS",
        "Mod_OnActvtItem": "ResRef,NSS",
        "Mod_OnClientEntr": "ResRef,NSS",
        "Mod_OnClientLeav": "ResRef,NSS",
#        "Mod_OnCutsnAbort": "ResRef,NSS",
        "Mod_OnHeartbeat": "ResRef,NSS",
        "Mod_OnModLoad": "ResRef,NSS",
        "Mod_OnModStart": "ResRef,NSS",
        "Mod_OnPlrDeath": "ResRef,NSS",
        "Mod_OnPlrDying": "ResRef,NSS",
#        "Mod_OnPlrEqItem": "ResRef,NSS",
        "Mod_OnPlrLvlUp": "ResRef,NSS",
        "Mod_OnPlrRest": "ResRef,NSS",
#        "Mod_OnPlrUnEqItm": "ResRef,NSS",
        "Mod_OnSpawnBtnDn": "ResRef,NSS",
#        "Mod_OnUnAcreItem": "ResRef,NSS",
        "Mod_OnUsrDefined": "ResRef,NSS",
        "Mod_StartDay": "Integer,1-31",
        "Mod_StartHour": "Integer,0-23",
        "Mod_StartMonth": "Integer,1-24",
        "Mod_StartYear": "Integer,0-2000",
        "Mod_Tag": "CExoString",
        "Mod_XPScale": "Integer,0-255"
        }
    def __init__(self,fname):
        NeverData.__init__(self)        
        self.erfFile = ERFFile()
        self.erfFile.fromFile(fname)
        ifoEntry = self.erfFile.getEntryByNameAndExtension("module","IFO")
        self.addPropList('ifo',self.ifoPropList,
                         self.erfFile.getEntryContents(ifoEntry).getRoot())

    def getProperty(self,label):
        '''overriden to only allow props in prop list'''
        if label in self.ifoPropList:
            return [label,
                    self.gffstructDict['ifo'].getInterpretedEntry(label),
                    self.ifoPropList[label]]

    def removeProperty(self,label):
        if label in self.ifoPropList:
            (s,t) = self.getTargetStructAndLabel(self.gffstructDict['ifo'],label)
            print 'removing',t
            s.removeEntry(t)
            
    def setProperty(self,label,value):
        if label in self.ifoPropList:
            (s,t) = self.getTargetStructAndLabel(self.gffstructDict['ifo'],label)
            if t == 'Mod_HakList':
                converted = []
                for n in value:
                    hak = GFFStruct(8)
                    hak.add('Mod_Hak',n,'CExoString')
                    converted.append(hak)
                value = converted
            s.setInterpretedEntry(t,value)

    def getName(self,lang=0):
        return self.gffstructDict['ifo'].getInterpretedEntry('Mod_Name').getString()
    
    def getAreaNames(self):
        """Get the names of areas in this ERF."""
        areas = self.erfFile.getEntriesWithExtension('ARE')
        names = []
        for area in areas:
            names.append(area.name)
        return names

    def getArea(self,name):
        '''Get an area by its name
        @param name name of the area object to return'''
        return Area(self.erfFile,name)
    
    def getAreas(self):
        """Get the areas in this ERF.
        @return a list of Area objects."""
        names = self.getAreaNames()
        areas = {}
        c = 1.0
        for n in names:
            self.setProgress((c/len(names))*100.0)
            areas[n] = self.getArea(n)
            c += 1
        self.setProgress(0)
        return areas

    def getScripts(self):
        """Get the scripts in this ERF.
        @return A list of Script objects."""
        entries = self.erfFile.getEntriesWithExtension('NSS')
        scripts = {}
        for s in entries:
            scripts[s.name] = self.erfFile.getEntryContents(s)
        return scripts

    def saveToReadFile(self):
        self.erfFile.saveToReadFile()

    def toFile(self,fpath):
        self.erfFile.toFile(fpath)

    def saveAs(self,fpath):
        self.erfFile.toFile(fpath)
        self.erfFile.reassignReadFile(fpath)
        
    def addERFFile(self,fname):
        self.erfFile.addFile(fname)
        
    def setProgressDisplay(self,p):
        self.erfFile.setProgressDisplay(p)

    def getEntriesWithExtension(self,ext):
        return self.erfFile.getEntriesWithExtension(ext)

    def setProgress(self,p):
        self.erfFile.setProgress(p)
