#!/usr/bin/python
# NOTE: Mac users, you need to run this as 'pythonw NeverEditMainApp.py'
# or change the line above to say 'pythonw'

import check_versions

import sys
if sys.version_info[0] < 2 or (sys.version_info[0] == 2 and sys.version_info[1] < 3):
    print >>sys.stderr,"Sorry, neveredit needs python >= 2.3. Download from python.org"
    sys.exit()

try:
    import wxPython
    from wxPython.wx import *
    import wx.html
    from wxPython.lib.buttons import wxGenButton
    from wxPython.lib.filebrowsebutton import DirBrowseButton
except:
    print >>sys.stderr,"Sorry,You don't seem to have a proper installation of wxPython."
    print >>sys.stderr,"Get one from wxpython.org."
    import traceback
    traceback.print_exc()
    sys.exit()

import Module
import Area
import MapWindow
import ResourceManager
import globals

import os
import threading
from sets import Set
import gettext

gettext.install('neveredit','translations')

ID_ABOUT = 101
ID_OPEN  = 102
ID_SAVE  = 103
ID_SAVEAS = 104
ID_EXIT  = 105
ID_HELP = 106
ID_PREFS = 107
ID_ADD_ERF = 108

ID_TIMER = 200

VERSION = 'v0.6'

def iAmOnMac():
    return sys.platform == 'darwin'

def iAmOnLinux():
    return 'linux' in sys.platform

def cleanstr(str):
    import string
    allchars = string.maketrans('','')
    delchars = ''.join([c for c in allchars if c not in string.printable])
    return str.translate(allchars,delchars)

class MySplashScreen(wx.SplashScreen):
    def __init__(self,pic):
        bmp = wx.Image(pic).ConvertToBitmap()
        wx.SplashScreen.__init__(self, bmp,
                                 wx.SPLASH_CENTRE_ON_SCREEN|wx.SPLASH_NO_TIMEOUT,
                                 4000, None, -1,
                                 style = wx.SIMPLE_BORDER|wx.FRAME_NO_TASKBAR|wx.STAY_ON_TOP)

##\mainpage
class NeverEditMainWindow(wxFrame):
    '''<html><body>

    <table width="100%" border=0>
    <tr>
    <td valign="top">
    <h2>neveredit</h2>
    </td><td align="right" valign="top">
    <img src="neveredit.jpg"></td></tr></table>

    <p>Welcome to neveredit. Neveredit strives to be an editor for files
    from the Bioware game Neverwinter Nights. One day it may have all of
    the functionality of the Bioware windows tools, and maybe more. For
    now, I am striving to achieve basic editing functionality on
    non-Windows platforms. This means that this is alpha quality
    software, and will at the current stage likely do bad things
    to your files. I am happy to receive bug reports, but take no
    responsibility for any damages incurred through the use of this
    software.</p>

    <p>I write neveredit as part of the <b>
    Openknights Consortium (http://sf.net/projects/openknights)</b>,
    a group of people dedicated to bringing a toolset for NWN to
    the Mac and to Linux (and maybe back to Windows). Please file any
    bug reports on our sourceforge site. To discuss this app, please
    use our sourceforge forums or the neveredit wiki page at
    http://openknights.sourceforge.net/tikiwiki/tiki-index.php?page=neveredit</p>
    
    <p>This is an application written in <b>Python
    (http://python.org)</b> and <b>wxPython
    (http://wxpython.org)</b>.</p>

    <p>Thanks to Damon Law (damonDesign.com) for the neveredit image.</p>

    <p>Thanks to Sylvan_Elf for the neveredit splash screen.</p>
    
    <p>Have fun,<br><i>sumpfork</i></p>

    </body></html>'''
    
    def __init__(self,parent,id,title):
        '''Constructor. Sets up static conrols and menus.'''
        wx.InitAllImageHandlers()
        self.splash = MySplashScreen("neveredit_logo.jpg")
        EVT_CLOSE(self.splash,self.OnCloseSplash)
        
        self.splash.Show(True)

        wxFrame.__init__(self,parent,-1,title,size=(600,600),\
                         style=wxDEFAULT_FRAME_STYLE\
                         |wxNO_FULL_REPAINT_ON_RESIZE)

        self.doInit = False
        self.fname = None
        self.doRead = False
        self.fileChanged = False
        self.propsChanged = False
        self.NWNAppDir = None
	self.NWNModDir = os.path.dirname(sys.argv[0])
	self.NWNERFDir = os.path.dirname(sys.argv[0])

        self.idToTreeItemMap = {}
        
        #status bar
        self.CreateStatusBar(2)
        self.SetStatusWidths([-1,150])
        self.SetStatusText(_("Welcome to neveredit..."))
        rect = self.GetStatusBar().GetFieldRect(1)
        self.statusProgress = wxGauge(self.GetStatusBar(),-1,100,
                                      wxPoint(rect.x+1,rect.y+1),
                                      wxSize(rect.width-2,rect.height-2))
        self.setProgress(0)
        
        if iAmOnMac():
            wx.App_SetMacExitMenuItemId(ID_EXIT)
            wx.App_SetMacPreferencesMenuItemId(ID_PREFS)
            wx.App_SetMacAboutMenuItemId(ID_ABOUT)

        #menus
        self.filemenu = wxMenu()
        self.filemenu.Append(ID_OPEN, '&' + _('Open') + '\tCtrl+O',
                             _("Open a File"))
        self.filemenu.Append(ID_ADD_ERF, _('Add in ERF File...'),
                             _("Add all entries of another ERF file"))
        self.filemenu.AppendSeparator()
        self.filemenu.Append(ID_SAVE, '&' + _('Save') + '\tCtrl+S',
                             _("Save File"))
        self.filemenu.Append(ID_SAVEAS, '&' + _('Save As...') +
                             '\tShift+Ctrl+S',
                             _("Save File under a new name"))
        if not iAmOnMac():
            self.filemenu.Append(ID_EXIT,_('E&xit') + '\tAlt-X',
                                 _("Quit neveredit"))
        self.filemenu.Enable(ID_SAVE,False)
        self.filemenu.Enable(ID_SAVEAS,False)
        self.filemenu.Enable(ID_ADD_ERF,False)

        self.filehistory = wxFileHistory()
        self.filehistory.UseMenu(self.filemenu)
        EVT_MENU_RANGE(self,wxID_FILE1,wxID_FILE9,self.OnFileHistory)
        
        editmenu = wxMenu()
        editmenu.Append(ID_PREFS,'&' + _('Preferences...'),
                        _('neveredit prefs'))
        
        helpmenu = wxMenu()
        helpmenu.Append(ID_ABOUT, '&' + _('About...'), _("About neveredit"))
        helpmenu.Append(ID_HELP,'&'+ _('No Help Yet'), _("neveredit Help"))
        
        menuBar = wxMenuBar()
        menuBar.Append(self.filemenu,"&" + _("File"))
        menuBar.Append(editmenu, "&" + _("Edit"))
        menuBar.Append(helpmenu, "&" + _("Help"))
        self.SetMenuBar(menuBar)
            
        EVT_MENU(self,ID_OPEN,self.openFile)
        EVT_MENU(self,ID_ADD_ERF,self.addERFFile)
        EVT_MENU(self,ID_SAVE,self.saveFile)
        EVT_MENU(self,ID_SAVEAS,self.saveFileAs)
        EVT_MENU(self,ID_ABOUT,self.about)
        EVT_MENU(self,ID_PREFS,self.prefs)
        EVT_MENU(self,ID_EXIT,self.exit)
        
        wxRegisterId(ID_OPEN)
        wxRegisterId(ID_ADD_ERF)
        wxRegisterId(ID_SAVE)
        wxRegisterId(ID_SAVEAS)
        wxRegisterId(ID_ABOUT)
        wxRegisterId(ID_EXIT)
        wxRegisterId(ID_HELP)
        wxRegisterId(ID_PREFS)

        splitter = wx.SplitterWindow(self,-1,style=wx.NO_3D|wx.SP_3D)
        #def EmptyHandler(evt): pass
        #wx.EVT_ERASE_BACKGROUND(splitter,EmptyHandler)
        
        tID = wxNewId()
        self.tree = wxTreeCtrl(splitter,tID,wxDefaultPosition,\
                               wxDefaultSize,\
                               wxTR_HAS_BUTTONS)
        self.selectedTreeItem = None
        EVT_TREE_SEL_CHANGED(self,tID,self.treeSelChanged)

        self.notebook = wx.Notebook(splitter,-1)#,style=wx.CLIP_CHILDREN)
        #EVT_NOTEBOOK_PAGE_CHANGED(self,self.notebook.GetId(),self.OnNotebookPageChanged)
        #EVT_ERASE_BACKGROUND(self.notebook,self.OnNotebookEraseBackground)
        
        splitter.SplitVertically(self.tree,self.notebook,180)
        splitter.SetMinimumPaneSize(50)

        self.welcome = wx.html.HtmlWindow(self.notebook,-1,(400,400))
        self.notebook.AddPage(self.welcome,"Welcome to neveredit")
        self.welcome.SetPage(self.__doc__)

        
        self.makePropPage()
        self.readingERF = False
        self.RMThread = None
        if 'HOME' in os.environ:
            self.prefPath = os.environ['HOME']
        else:
            self.prefPath = os.path.dirname(sys.argv[0])
            
        if iAmOnMac():
            self.prefPath += '/Library/Preferences/org.openknights.neveredit.plist'
        elif sys.platform.find('linux') >= 0:
            self.prefPath += '/.neveredit'
        else:
            self.prefPath += '/neveredit.prefs'
            
        self.loadPrefs()

        self.timer = wxTimer(self,ID_TIMER)
        EVT_TIMER(self,ID_TIMER,self.kick)
        self.timer.Start(200)
        
        #self.Show(true)
        EVT_IDLE(self,self.idle)
        EVT_CLOSE(self,self.OnClose)

        tmp = self.splash
        self.splash = MySplashScreen("neveredit_logo_init.jpg")
        self.splash.Show(True)
        if tmp:
            tmp.Show(False)
            tmp.Destroy()
        
    def initResourceManager(self):
        '''Initialize the resource manager object from the app dir path.
        Fails with a dialog warning if path does not exit. If it does
        exist, disables interface and starts the initialization thread.'''
        if not os.path.exists(self.NWNAppDir):
            if self.splash:
                self.splash.Show(False)
                self.splash.Destroy()
            dlg = wxMessageDialog(self,_('Directory does not exist: ')
                                  + self.NWNAppDir +
                                  '\n' + _('Please reset in Preferences'),
                                  _('Non-existent App Dir'),wxOK|wxICON_ERROR)
            dlg.ShowModal()
            dlg.Destroy()
            self.Show(True)
        else:
            self.Enable(False)
            self.SetStatusText(_("Initializing from NWN Application Directory..."))
            self.RMThread = threading.Thread(target=self.doInitResourceManager)
            self.RMThread.start()

    def doInitResourceManager(self):
        '''This is the thread body for initializing the resource manager.Do
        not call any function that
        changes the interface from this method. You need to set a flag and call
        it from the main app thread, as otherwise things will get screwed up on
        linux.'''
        self.resourceManager = ResourceManager.ResourceManager()
        self.resourceManager.setProgressDisplay(self)
        self.resourceManager.scanGameDir(self.NWNAppDir)
        globals.resourceManager = self.resourceManager
        
    def setProgress(self,prog):
        '''Implementing the progress display interface'''
        self.progress = prog
    
    def makePropPage(self):
        '''Initialize the static controls for the properties notebook page.'''
        self.props = wxScrolledWindow(self.notebook,-1)
        self.propGrid = wxFlexGridSizer(0,2,7,7)
        self.propGrid.AddGrowableCol(0)
        self.propGrid.AddGrowableCol(1)
        self.props.SetSizer(self.propGrid)
        self.props.SetAutoLayout(True)
        self.props.Move((50,10))
        self.notebook.AddPage(self.props,_("Properties"))

    def treeFromERF(self):
        '''Make a new tree control from the erf file currently
        associated with the app. Does so in a new thread.'''
        self.selectedTreeItem = None
        self.cleanPropPage()
        self.tree.DeleteAllItems()
        self.treeRoot = self.tree.AddRoot(self.module.getName())
        self.tree.SetPyData(self.treeRoot,self.module)
        self.erfThread = threading.Thread(target=self.doTreeFromERF)
        wx.BeginBusyCursor()
        self.erfThread.start()
        self.readingERF = True

    def doTreeFromERF(self):
        '''The thread body for loading an ERF file.
        Do not call any function that changes the interface
        from this method. You need to set a flag and call
        it from the main app thread, as otherwise things will get screwed up on
        linux.'''
        self.module.setProgressDisplay(self)
        self.areas = self.module.getAreas()

    def setFileChanged(self,fc):
        '''Call this when the loaded file has changed
        (someone used a control, usually).'''
        self.fileChanged = fc
        self.filemenu.Enable(ID_SAVE,fc)
        
    def treeFromERFDone(self):
        '''This gets called when we finish loading a new ERF file.
        It gets the interface into its final state and cleans up a bit.'''
        areaRoot = self.tree.AppendItem(self.treeRoot,'Areas')
        for area in self.areas.values():
            self.subtreeFromArea(areaRoot,area)
        scriptRoot = self.tree.AppendItem(self.treeRoot,_('Scripts'))
        self.addScripts(scriptRoot)
        self.notebook.Refresh()
        self.SetStatusText(_("Read ") + self.fname)
        self.filemenu.Enable(ID_SAVEAS,True)
        self.filemenu.Enable(ID_ADD_ERF,True)
        self.fileChanged = False
        self.tree.Expand(self.treeRoot)
        self.tree.UnselectAll()
        self.tree.SelectItem(self.treeRoot)
        wx.EndBusyCursor()

    def OnFileHistory(self,event):
        '''Callback for someone selecting a file history menu item.'''
        fileNum = event.GetId() - wxID_FILE1
        self.readFile(self.filehistory.GetHistoryFile(fileNum))
        
    def OnNotebookPageChanged(self,event):
        '''Callback for notebook page change event'''
        if event.GetOldSelection() == 2:
            self.map.Show(False)
            size = self.notebook.GetPage(event.GetSelection()).GetRect()
            print 'refreshing',size
            self.notebook.GetPage(event.GetSelection()).Refresh(True,size)
        if event.GetSelection() == 2:
            self.map.Show(True)

    def addScripts(self,scriptRoot):
        '''Add the scripts contained in our module to the interface.'''
        scripts = self.module.getScripts()
        scriptNames = scripts.keys()
        scriptNames.sort()
        for s in scriptNames:
            name = s.split('.')[0]
            scriptItem = self.tree.AppendItem(scriptRoot,name)
            scripts[s].setNWNDir(self.NWNAppDir)
            self.tree.SetPyData(scriptItem,scripts[s])

    def init(self):
        '''Schedule the the app init routine.'''
        self.doInit = True

    def kick(self,event):
        '''kick the idle func even without events'''
        wxPython.wx.wxWakeUpIdle()
        
    def idle(self,event):
        '''Called on idle events. This is where the interface updates for
        other threads happen (they cannot update the interface themselves).
        For example, this handles updates to the progress bar and
        finishing up ERF file loading.'''
        if self.doInit:
            self.doInit = False
            self.initResourceManager()
        if self.RMThread and not self.RMThread.isAlive() and self.doRead:
            self.readFile(self.fname)
            self.doRead = False
        if self.RMThread and not self.RMThread.isAlive():
            if self.splash:
                self.splash.Show(False)
                self.splash.Destroy()
            self.Show(True)
            self.Enable(True)
            self.SetStatusText(_("Welcome to neveredit..."))
            self.RMThread = None
        #when we're finished reading an ERF, display it
        if self.readingERF and not self.erfThread.isAlive():
            self.readingERF = False
            self.treeFromERFDone()
        if self.progress > 0:
            self.setProgress(self.progress);
            self.statusProgress.Show(True)
            self.statusProgress.SetValue(int(self.progress))
        else:
            self.statusProgress.Show(False)

    def selectTreeItemById(self,oid):
        '''try to find an item in the current module by object id and
        select the corresponding tree item'''
        item = self.idToTreeItemMap.get(oid,None)
        if item:
            self.tree.EnsureVisible(item)
            self.tree.SelectItem(item)

    def OnMapSelection(self,event):
        '''handle a map window selection event'''
        self.selectTreeItemById(event.getSelectedId())

    def OnMapMove(self,event):
        '''handle a map movement event'''
        self.setFileChanged(True)
        
    def subtreeFromArea(self,item,area):
        '''Create a new subtree for a module area.'''
        areaItem = self.tree.AppendItem(item,area.getName())
        self.tree.SetPyData(areaItem,area)

        doors = area.getDoors()
        if len(doors) > 0:
            doorParentItem = self.tree.AppendItem(areaItem,_('Doors'))
            for door in doors:
                doorItem = self.tree.AppendItem(doorParentItem,
                                                door.getName())
                self.tree.SetPyData(doorItem,door)
                self.idToTreeItemMap[door.getNevereditId()] = doorItem
            
        placeables = area.getPlaceables()
        if len(placeables) > 0:
            placeableParentItem = self.tree.AppendItem(areaItem,
                                                       _('Placeables'))
            for placeable in placeables:
                placeableItem = self.tree.AppendItem(placeableParentItem,
                                                     placeable.getName())
                self.tree.SetPyData(placeableItem,placeable)
                self.idToTreeItemMap[placeable.getNevereditId()] = placeableItem

        creatures = area.getCreatures()
        if len(creatures) > 0:
            creatureParentItem = self.tree.AppendItem(areaItem,
                                                      _('Creatures'))
            for creature in creatures:
                creatureItem = self.tree.AppendItem(creatureParentItem,
                                                    creature.getName())
                self.tree.SetPyData(creatureItem,creature)
                self.idToTreeItemMap[creature.getNevereditId()] = creatureItem
        
        items = area.getItems()
        if len(items) > 0:
            itemParentItem = self.tree.AppendItem(areaItem,_('Items'))
            for item in items:
                itemItem = self.tree.AppendItem(itemParentItem,
                                                item.getName())
                self.tree.SetPyData(itemItem,item)
                self.idToTreeItemMap[item.getNevereditId()] = itemItem
        
    def getAreaForTreeItem(self,item):
        '''Get the area associated with this tree item'''
        while item:
            data = self.tree.GetPyData(item)
            if data and data.__class__ == Area.Area:
                return data
            else:
                item = self.tree.GetItemParent(item)
        return None
    
    def treeSelChanged(self,event):
        '''Callback to handle the user changing the selection
        in the main tree.'''
        self.maybeApplyPropControlValues()
        self.selectedTreeItem = event.GetItem()
        data = self.tree.GetPyData(self.selectedTreeItem)
        #self.notebook.SetSelection(0)
        if data:
            self.makePropsForItem(data)
            area = self.getAreaForTreeItem(self.selectedTreeItem)
            if area and self.notebook.GetPageCount() == 2:
                self.map = MapWindow.MapWindow(self.notebook)
                self.map.setArea(area)
                MapWindow.EVT_MAPSINGLESELECTION(self.map,self.OnMapSelection)
                MapWindow.EVT_MAPMOVE(self.map,self.OnMapMove)                
                self.notebook.AddPage(self.map,_('Map'))
                self.notebook.SetSelection(2)
                self.map.selectThingById(data.getNevereditId())
            elif area:
                self.map.setArea(area)
                self.notebook.SetSelection(2)
                self.map.selectThingById(data.getNevereditId())
            elif self.notebook.GetPageCount() == 3:                
                self.map.setArea(None)
                if not iAmOnLinux():
                    self.notebook.DeletePage(2)
                self.notebook.SetSelection(1)
            else:
                self.notebook.SetSelection(1)
        else:
            self.cleanPropPage()

    propControls = {}
    propLabels = []

    def maybeApplyPropControlValues(self):
        if self.selectedTreeItem:
            data = self.tree.GetPyData(self.selectedTreeItem)
            if data:
                self.applyPropControlValues()

    def applyPropControlValues(self):
        '''This method reads back in the valus of currently displayed property controls
        and updates the actual module file to reflect these values.'''
        if not self.propsChanged:
            return
        print 'applying prop changes now...'
        for propControl in self.propControls.values():
            control = propControl[0]
            prop = propControl[1]
            typeSpec = prop[2].split(',')
            pName = typeSpec[0]
            if pName == 'CExoLocString':
                prop[1].setString(control.GetValue())
            elif pName == 'CExoString':
                prop[1] = control.GetValue()
            elif pName == 'Percentage':
                prop[1] = control.GetValue()
            elif pName == 'Boolean':
                prop[1] = int(control.GetValue())
            elif pName == 'Integer':
                prop[1] = int(control.GetValue())
            elif pName == 'BGRColour':
                c = control.GetBackgroundColour()
                prop[1] = (c.Blue() << 16) | (c.Green() << 8) | c.Red()
            elif pName == 'ResRef':
                prop[1] = control.GetStringSelection()
            elif pName == 'Script':
                s = control.GetValue()
                prop[1] = s.replace('\n','\r\n')
            elif pName == 'CheckList':
                type = prop[2].split(',')[1]
                checkedList = []
                if type == 'HAKs':
                    for i in range(control.GetCount()):
                        if control.IsChecked(i):
                            hakFileName = control.GetString(i)
                            checkedList.append(hakFileName)
                prop[1] = checkedList
                print 'setting',prop[0]
            elif pName == '2daIndex':
                prop[1] = int(control.GetSelection())
            else:
                print _('error, unknown prop type:'),pName
            #print 'set',pName,prop[0],'to',prop[1]
            self.tree.GetPyData(self.selectedTreeItem).setProperty(prop[0],prop[1])
        self.propsChanged = False
            
    def makeControlForProp(self,prop,parent):
        '''Make a wxWindows control for the given NWN property and add
        it to the given parent wxWindow.'''
        label = wxStaticText(self.props,-1,prop[0].split('.')[-1],
                             wxDefaultPosition,(100,20))
        control = None
        typeSpec = prop[2].split(',')
        type = typeSpec[0]
        if type == 'CExoLocString':
            if len(typeSpec) > 1 and int(typeSpec[1]) > 1:
                control = wxTextCtrl(parent,-1,prop[1].getString(),
                                     wxDefaultPosition,
                                     (200,int(typeSpec[1])*24),
                                     style=wxTE_MULTILINE)
            else:
                control = wxTextCtrl(parent,-1,prop[1].getString(),
                                     wxDefaultPosition,
                                     (200,24))                
            EVT_TEXT(self.props,control.GetId(),self.controlUsed)
        elif type == 'CExoString':
            control = wxTextCtrl(parent,-1,prop[1],wxDefaultPosition,(200,24))
            EVT_TEXT(self.props,control.GetId(),self.controlUsed)
        elif type == 'Percentage':
            control = wxSlider(parent,-1,prop[1],0,100,wxDefaultPosition,
                               wxSize(200,20),
                               wxSL_HORIZONTAL|wxSL_LABELS)
            EVT_COMMAND_SCROLL(self.props,control.GetId(),self.controlUsed)
            #control.SetTickFreq(5,0)
        elif type == 'Boolean':
            control = wxCheckBox(parent,-1,'')
            control.SetValue(prop[1])
            EVT_CHECKBOX(self.props,control.GetId(),self.controlUsed)
        elif type == 'Integer':
            min = 0
            max = 100
            if len(typeSpec) > 1:
                maxMin = typeSpec[1].split('-')
                min = int(maxMin[0])
                max = int(maxMin[1])
            control = wxSlider(parent,-1,prop[1],min,max,wxDefaultPosition,
                               wxSize(200,20),
                               wxSL_HORIZONTAL|wxSL_AUTOTICKS|wxSL_LABELS)
            EVT_COMMAND_SCROLL(self.props,control.GetId(),self.controlUsed)
            #control.setTickFreq(10,0)
        elif type == "ResRef":
            if len(typeSpec) > 1:
                entries = self.module.getEntriesWithExtension(typeSpec[1])
                keyList = [''] #can leave empty
                keyList.extend([x.name for x in entries])
                selection = prop[1]
                try:
                    index = keyList.index(selection)
                except ValueError:
                    keyList.append(selection)
                    index = len(keyList) - 1
                control = wxChoice(parent,-1,choices=keyList)                
                if len(prop[1]) > 0:
                        control.SetSelection(index)
                else:
                    control.SetSelection(0)
                EVT_CHOICE(self.props,control.GetId(),self.controlUsed)
            else:
                control = wxTextCtrl(parent,-1,prop[1])
                EVT_TEXT(self.props,control.GetId(),self.controlUsed)
        elif type == "BGRColour":
            blue = prop[1] >> 16
            green = (prop[1] >> 8) & (0xff)
            red = prop[1] & (0xff)            
            control = wxGenButton(parent,-1,'',wxDefaultPosition,
                                  wxSize(40,40))
            control.SetBezelWidth(0)
            control.SetForegroundColour(wxColour(red,green,blue))
            control.SetBackgroundColour(wxColour(red,green,blue))
            EVT_BUTTON(self,control.GetId(),self.handleColourButton)
        elif type == "CheckList":
            choices = Set()
            if typeSpec[1] == 'HAKs':
                choices.update([x.split('.')[0] for x in
                                self.resourceManager.getHAKFileNames()])
                choices.update([x.getInterpretedEntry('Mod_Hak')
                                for x in prop[1]])
            elif typeSpec[1] == 'HAK': #old format
                choices.update([x.split('.')[0] for x in
                                self.resourceManager.getHAKFileNames()])
                if len(prop[1]) > 0:
                    choices.add(prop[1])
                #update the module to the new haklist format
                prop[2] = 'CheckList,HAKs'
                prop[0] = 'Mod_HakList'
                self.setFileChanged(True)
                self.propsChanged = True
                self.SetStatusText(_("Old Module, changing to new format Mod_HakList..."))
            choices = list(choices)
            choices.sort()
            control = wxCheckListBox(parent,-1,
                                     wxDefaultPosition,(200,200),choices)
            for i in range(len(choices)):
                control.Check(i,False)
            if typeSpec[1] == 'HAK':
                if len(prop[1]) > 0:
                    control.Check(choices.index(prop[1]))
            else:
                for n in prop[1]:
                    control.Check(choices.index(n.getInterpretedEntry('Mod_Hak')))
            EVT_CHECKLISTBOX(self,control.GetId(),self.controlUsed)
        elif type == '2daIndex':
            twoda = self.resourceManager.getResource(ResourceManager.keyFromName(typeSpec[1]))
            choices = []
            col = typeSpec[2]
            if typeSpec[3] == 'strref':
                for i in xrange(twoda.getRowCount()):
                    entry = 'invalid'
                    try:
                        entry = self.resourceManager.getDialogString(int(twoda.getEntry(i,col)))
                    except ValueError:
                        if len(typeSpec) > 4:
                            entry = twoda.getEntry(i,typeSpec[4])
                    choices.append(entry)
            else:
                choices = [twoda.getEntry(i,col) for i in xrange(twoda.getRowCount())]
            control = wxChoice(parent,-1,choices=[cleanstr(s) for s in choices])
            control.SetSelection(prop[1])
            EVT_CHOICE(self.props,control.GetId(),self.controlUsed)
        return (label,control)

    def handleColourButton(self,event):
        '''Callback for the colour selection dialog.'''
        control = self.propControls[event.GetId()]
        cd = wxColourData()
        cd.SetColour(control[0].GetBackgroundColour())
        dlg = wxColourDialog(self.props,cd)
        if dlg.ShowModal() == wxID_OK:
            cd = dlg.GetColourData()
            control[0].SetBackgroundColour(cd.GetColour())
            self.controlUsed(event)
            
    def cleanPropPage(self):
        '''Clean up the prop page, removing all labels and controls.'''
        for l in self.propLabels:
            self.propGrid.Remove(l)
            l.Destroy()
        for c,p in self.propControls.values():
            self.propGrid.Remove(c)
            #if not iAmOnMac() or c.__class__ != wxCheckListBox:
            c.Destroy()
            #else:
                #this appears to be a bug? Destroy() crashes on Mac.
                #c.GetParent().RemoveChild(c)
        self.propControls = {}
        self.propLabels = []
        self.propGrid.SetCols(2)
        self.notebook.Refresh()

    def makePropsForItem(self,item):
        '''Make all property controls for a given item.
        The item must be implementing the NeverData interface.'''
        self.cleanPropPage()
        needScrollbars = False
        for p in item.iterateProperties():
            if p[1] == None:
                print 'empty prop for',p[0]
                continue
            if p[0] == 'Mod_Hak':
                item.removeProperty(p[0])
            if p[2] == 'Script':
                #I tried wxEditor, but it's not good on Mac
                #editor = NeverEditor(self.props,-1)
                #editor.SetText(p[1].split('\n'))
                s = p[1].replace('\r','')
                control = wxTextCtrl(self.props,-1,s,style=wxTE_MULTILINE)
                self.propGrid.SetCols(1)
                self.propGrid.Add(control,1,wxALL|wxEXPAND,5)
                self.propGrid.AddGrowableRow(0)
                EVT_TEXT(self.props,control.GetId(),self.controlUsed)
            else:
                (label,control) = self.makeControlForProp(p,self.props)
                if control:
                    self.propGrid.Add(label,1,
                                      wxALIGN_CENTRE|wxALIGN_CENTRE_VERTICAL|wxALL,0)
                    self.propLabels.append(label)
                    self.propGrid.Add(control,1,
                                      wxALIGN_CENTRE|wxALIGN_CENTRE_HORIZONTAL|wxALL,5)
                    needScrollbars = True
                else:
                    print 'Error: unhandled prop type',p[2]
                    if label:
                        label.Destroy()
            if control:
                self.propControls[control.GetId()] = (control,p)
        self.propGrid.Layout()
        #all this doesn't seem to work on wxMac, so do it manually
        if needScrollbars:
            self.propGrid.FitInside(self.props)
            size = self.propGrid.GetSize()
            self.props.SetScrollbars(20,20,size.x/20,size.y/20)
        else:
            self.props.SetScrollbars(0,0,0,0)
        self.props.Refresh()
        self.notebook.Refresh()

    def controlUsed(self,event):
        '''Callback for a control being used. Sets a dirty flag.'''
        self.setFileChanged(True)
        self.propsChanged = True
        
    def about(self,event):
        '''About menu item callback.'''
        dlg = wxMessageDialog(self,_('neveredit ') + VERSION +
                              _(''' by Sumpfork
The OpenKnights Consortium
Copyright 2003-2004'''),
                                     _('About neveredit'),
                                     wxOK | wxICON_INFORMATION)
        dlg.ShowModal()
        dlg.Destroy()

    def maybeSave(self):
        '''Ask whether we should save changes and do so if yes.
        @return boolean indicating whether changes were saved.'''
        if self.fileChanged:
            dlg = wxMessageDialog(self,_('Save Changes to ') + self.fname
                                  + '?',
                                  _("Changed File"),wxYES_NO|wxICON_QUESTION)
            if dlg.ShowModal() == wxID_YES:
                self.saveFile(None)
                return True
            else:
                return False
        else:
            return True

    def askDiscard(self,msg):
        '''Ask whether we should discard unsaved changes.
        @param msg the message to use in the dialog
        @return boolean indicating whether the use said we should
        discard (True)
        '''
        dlg = wxMessageDialog(self,msg,
                              _("Unsaved Changes"),wxYES_NO|wxICON_QUESTION)
        return dlg.ShowModal() == wxID_YES

    def OnCloseSplash(self,event):
        self.splash = None
        
    def OnClose(self,doForce=False):
        '''Window closing callback method for the main app.'''
        self.savePrefs()
        if self.maybeSave():
            sys.exit()
        else:
            if self.askDiscard(_('Discard unsaved changes and exit neveredit?')):
                sys.exit()
            else:
                return False
        
    def exit(self,event):
        '''Exit the Main app, asking about possible unsaved changes.'''
        if self.OnClose():
            sys.exit()
        
    def prefs(self,event):
        '''Display a prefs dialog.'''
        dlg = wxDialog(self,-1,_("Preferences"),
                       wxDefaultPosition,wxSize(500,90))
        sizer = wxBoxSizer(wxVERTICAL)
        appDirButton = DirBrowseButton(dlg,-1,wxDefaultPosition,(500,30),
                              labelText=_('NWN Directory'),
                              buttonText=_('Select...'),
                              startDirectory=self.NWNAppDir)
        appDirButton.SetValue(self.NWNAppDir)
        sizer.Add(appDirButton,0,wxGROW|wxALIGN_CENTRE|wxALL,5)

	modDirButton = DirBrowseButton(dlg,-1,wxDefaultPosition,(500,30),
	                      labelText=_('Modules Directory'),
			      buttonText=_('Select...'),
			      startDirectory=self.NWNModDir)
        modDirButton.SetValue(self.NWNModDir)
	sizer.Add(modDirButton,0,wxGROW|wxALIGN_CENTRE|wxALL,5)

	erfDirButton = DirBrowseButton(dlg,-1,wxDefaultPosition,(500,30),
	                      labelText=_('ERFs Directory'),
			      buttonText=_('Select...'),
			      startDirectory=self.NWNERFDir)
	erfDirButton.SetValue(self.NWNERFDir)
	sizer.Add(erfDirButton,0,wxGROW|wxALIGN_CENTRE|wxALL,5)

        line = wxStaticLine(dlg,-1,size=(20,-1),style=wxLI_HORIZONTAL)
        sizer.Add(line,0,wxGROW|wxALIGN_CENTER_VERTICAL|wxRIGHT|wxTOP,5)
        
        box = wxBoxSizer(wxHORIZONTAL)
        
        btn = wxButton(dlg,wxID_CANCEL,_("Cancel"))
        box.Add(btn,0,wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL|wxALL,5)

        btn = wxButton(dlg,wxID_OK,_("OK"))
        btn.SetDefault()
        box.Add(btn,0,wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL|wxALL,5)

        sizer.AddSizer(box,0,wxALIGN_RIGHT|wxALL,5)

        dlg.SetSizer(sizer)
        dlg.SetAutoLayout(True)
        sizer.Fit(dlg)
        dlg.Layout()
        dlg.CenterOnParent()
        
        if dlg.ShowModal() == wxID_OK:
            self.NWNAppDir = appDirButton.GetValue()
	    self.NWNModDir = modDirButton.GetValue()
	    self.NWNERFDir = erfDirButton.GetValue()
            print 'app dir now',self.NWNAppDir
	    self.initResourceManager()
            self.savePrefs()

    def savePrefs(self):
        '''Save the current preferences settings.'''
        import codecs,plistlib,encodings        
        codecs.register(encodings.search_function)
        n = self.filehistory.GetCount()
        files = []
        for i in range(n):
            files.append(self.filehistory.GetHistoryFile(i))
        pl = plistlib.Plist(NWNAppDir=self.NWNAppDir,
                            NWNModDir=self.NWNModDir,
			    NWNERFDir=self.NWNERFDir,
			    FileHistory=files)
        pl.write(self.prefPath)

    def loadPrefs(self):
        '''Load preferences from their standard location.'''
        #defaults
        if not self.NWNAppDir:
            if sys.platform == 'darwin':
                self.NWNAppDir = '/Applications/Neverwinter Nights/'
            elif sys.platform.find('linux') >= 0:
                self.NWNAppDir = '/usr/local/games/nwn/'
            else:
                self.NWNAppDir = '/Program Files/NWN/'                
        import codecs,plistlib,encodings
        codecs.register(encodings.search_function)
        if os.path.exists(self.prefPath):
            pl = plistlib.Plist.fromFile(self.prefPath)
            try:
                self.NWNAppDir = pl.NWNAppDir
		self.NWNModDir = pl.NWNModDir
		self.NWNERFDir = pl.NWNERFDir
                files = pl.FileHistory
                for f in files:
                    self.filehistory.AddFileToHistory(f)
            except:
                #could be out of date or malformed
                print sys.exc_info()[1] #fail silently

    def doReadFile(self,fname):
        '''Read in a file. Does not ask about unsaved changes.
        @param fname name of file to read
        '''
        self.SetStatusText("Reading " + fname + "...")
        self.module = Module.Module(fname)
        self.fname = fname
        self.treeFromERF()
        self.setFileChanged(False)
        self.filehistory.AddFileToHistory(fname)
        
    def readFile(self,fname):
        '''Read in a file. Asks about unsaved changes.
        @param fname name of file to read
        '''
        if self.maybeSave():
            self.doReadFile(fname)
        else:
            if self.askDiscard('Discard unsaved changes and load new file?'):
                self.doReadFile(fname)
        
        
    def openFile(self,event):
        '''Display a dialog to find a file, and load it if user says yes.'''
        dlg = wxFileDialog(self,_("Choose an ERF (mod/hak/nwm) File"),\
                           self.NWNModDir, '',\
                           'MOD|*.mod|ERF|*.erf|HAK|*.hak|SAV|*.sav|'+_('All Files')
                           +'|*.*',
                           wxOPEN)
        if dlg.ShowModal() == wxID_OK:
            self.readFile(dlg.GetPath())
        dlg.Destroy()

    def addERFFile(self,event):
        '''Display a dialog to find a file and add its entries to
        the current one.'''
        dlg = wxMessageDialog(self,_(
'Merging in a an ERF will overwrite the file you have currently\n\
loaded and save any changes you have made so far. Proceed?'),
                              _("Merge ERF File"),wxYES_NO|wxICON_QUESTION)
        if dlg.ShowModal() == wxID_NO:
            return
        dlg = wxFileDialog(self,_("Choose an ERF (mod/hak/nwm) File to add"),\
                           self.NWNERFDir, '',\
                           'ERF|*.erf|MOD|*.mod|HAK|*.hak|SAV|*.sav|'+_('All Files')
                           +'|*.*',
                           wxOPEN)
        if dlg.ShowModal() == wxID_OK:
            self.maybeApplyPropControlValues()
            self.notebook.SetSelection(0)
            self.tree.SelectItem(self.treeRoot)
            self.module.addERFFile(dlg.GetPath())
            self.treeFromERF()
        dlg.Destroy()
        
    def saveFile(self,event):
        '''Save the file to the file name we loaded it from.'''
        self.maybeApplyPropControlValues()
        self.module.saveToReadFile()
        self.setFileChanged(False)

        
    def saveFileAs(self,event):
        '''Save the file to a filename the users specifies in a file dialog.'''
        dlg = wxFileDialog(self,_("Choose a an ERF (mod/hak/nwm) File Name for Saving"),
                           os.getcwd(), '',\
                           'MOD|*.mod|HA|*.hak|'+_('All Files') + '|*.*',
                           wxSAVE)
        if dlg.ShowModal() == wxID_OK:
            self.maybeApplyPropControlValues()
            self.module.saveAs(dlg.GetPath())
            self.fname = dlg.GetPath()
        dlg.Destroy()
        self.setFileChanged(False)

def run():
    app = wxPySimpleApp()
    frame = NeverEditMainWindow(None,-1,_("neveredit"))
    if(len(sys.argv) > 1 and sys.argv[1][0] != '-'):
        frame.NWNAppDir = sys.argv[1]
    if(len(sys.argv) > 2):
        frame.fname = sys.argv[2]
        frame.doRead = True
    frame.init()
    app.MainLoop()
    
if __name__ == "__main__":
    #sys.stderr = wxPyNonFatalErrorDialog(None,-1,programname='neveredit',
    #                                     mailto='sumpfork@users.sourceforge.net',
    #                                     version='0.3',
    #                                     whendismissed='run()')
    if len(os.path.dirname(sys.argv[0])) > 0:
        # mainly to find neveredit.jpg in mac os app bundle
        os.chdir(os.path.dirname(sys.argv[0])) 
    run()
