#!/usr/bin/env python

import sys
import ResourceManager
from optparse import OptionParser
from ERFFile import ERFFile

if __name__ == "__main__":
    try:
        parser = OptionParser(usage='usage: %prog [options] <lookup|get|print> args')
        parser.add_option('-d','--nwndir',type="string",dest="appDir",
                          help="NWN Application Dir",
                          default="/Applications/Neverwinter Nights/")
        parser.add_option('-m','--module',type='string',dest='module',
                          help='Module file to read')
        parser.add_option('-o','--out',type='string',dest='outfile',
                          help='output file (stdout by if not specified)')
        (options,args) = parser.parse_args()
        outfile = sys.stdout
        if(len(args) == 0):
            print '''possible commands are: lookup get print
            "lookup" takes a list of extensions and prints all known
            resources with those extensions
            "get" retrieves a resource in raw format
            "print" prints a resource in text format (if possible)
            "strref" looks up and prints a string ref in dialog.tlk (hex allowed))
            use the "-h" switch to get a list of options'''
        else:
            command = args[0]
            if command == 'lookup':
                if len(args) == 1:
                    print 'lookup requires a list of extensions'
                else:
                    r = ResourceManager.ResourceManager()
                    r.scanGameDir(options.appDir)
                    if options.module:
                        r.addMODFile(options.module)
                    for e in args[1:]:
                        print >> sys.stderr,'resources with extension "' + e.upper() + '":'
                        for k in r.getKeysWithExtensions(e.upper()):
                            print >> outfile,ResourceManager.keyToString(k)
            elif command == 'get' or command == 'print':
                if len(args) == 1:
                    print 'require key in form resref.ext'
                else:
                    r = ResourceManager.ResourceManager()
                    r.scanGameDir(options.appDir)
                    if options.module:
                        print >>sys.stderr,'adding module',options.module
                        r.addMODFile(options.module)
                    if command == 'print':
                        if options.outfile:
                            outfile = open(options.outfile,'w')
                        print >> outfile,r.getResource(ResourceManager.keyFromName(args[1]))
                    else:
                        if options.outfile:
                            outfile = open(options.outfile,'wb')
                        outfile.write(r.getRawResource(ResourceManager.keyFromName(args[1])))
            elif command == 'strref':
                if len(args) == 1:
                    print 'require strref as integer'
                else:
                    r = ResourceManager.ResourceManager()
                    r.scanGameDir(options.appDir)
                    for numstr in args[1:]:
                        num = -1
                        if numstr.startswith('0x'):
                            num = int(numstr[2:],16)
                        else:
                            num = int(numstr)
                        print >> outfile,r.getDialogString(num)
    except KeyboardInterrupt:
        print >>sys.stderr,'\naborting...'
